## How ActiveInbox is a Local Client App (no Gmail user data reaches ActiveInbox's servers)

ActiveInbox is a local client app in that no Gmail user data - not even OAuth tokens - are transmitted to any of ActiveInbox's servers. All communication that includes Gmail data occurs only between the local client and Google's servers. 

Here are the actions we took:

**No OAuth Tokens Pass Through ActiveInbox's Servers (purely app-client<-->Google)**

(For full details on this, see README_GOOGLE_OAUTH.MD). 

The OAuth flow is modelled on https://developers.google.com/identity/protocols/OAuth2InstalledApp. 

Key Decisions
* See the code module webAuthFlowDirectOffline for how authentication occurs. 
* It uses chrome.identity.launchWebAuthFlow, with PKCE and state to prevent MITM, to authenticate directly with Google's servers from the client. 
* The OAuth Refresh token never escapes from a single closure/module, kept in the Background page, so even if ActiveInbox is compromised the malicious code cannot easily access it. 
* The OAuth Refresh token is stored on the hard drive only after being split into two pieces using a Vernam Cipher/One Time Pad. One piece is stored in the Extension Background Page's document.cookie (further encrypted by Chrome on disk), and the other in the Extension's Storage. 

Controversial Aspects
* We chose to use chrome.identity.launchWebAuthFlow over chrome.identity.getAuthToken, because our customers are power users (as most Extension users are), and typically have multiple Gmail accounts used in one Chrome window/profile (whereas chrome.identity.getAuthToken forces one Gmail account per Chrome window, which is a difficult limitation to explain without complex UX and forced behaviours). 
* In the above link (OAuth2InstalledApp), it's accepted that the Client Secret is no longer a secret (as it's stored in app client code), but Google is at least notified of that by using a GCP OAuth Application Type of iOS/Android. However, as we use chrome.identity.launchWebAuthFlow (see above), we have an Application Type of 'Web Application', where it's not as clear the Client Secret is exposed. 

**All Data Sent To ActiveInbox's Server Is Validated, And Checked To Prove No OAuth Tokens In It**

The aim is to prevent any Gmail User Data accidentally being sent to our server. The only acceptable data is preferences, notes, sub tasks, and stats. 

The default module function activeInboxAPI is responsible for every request to the ActiveInbox server. It does two things to ensure no sensitive data is uploaded:
* Any URL or body in the request, immediately pre-send, is passed through validateStringContainsNoKnownTokens. This asks the Auth module to test the request data against any known (only within itself) OAuth Tokens, to ensure none are present. 
* If the request is a POST/PATCH, then the body *must* pass through a validator that has been set at run time (i.e. validators cannot be added later, so no malicious code can create them). The validator tests that the body conforms to a Yup schema. These schemas are easy to read (e.g. activeInboxPrefsSchema). The principle goal of the schema is to have no long strings, and no objects with unknown keys - i.e. no way to slip in Gmail User Data. 



## Extension Security

* Runs Javascript in strict mode, to reduce unexpected side effects. 
* Tightly locked down CSP with default-src set to 'none', so if compromised by malicious code, limits which servers can be reached. 
* Modified the Window/Document/Element objects in the content script to disable innerHTML, and prevent the creation of SCRIPT tags, to stop accidental execution of code from (potentially malicious) strings. 
* All requests to Authenticate, Gmail, GCal, or ActiveInbox server goes through background page. Background page validates that data before communicating with a remote server. 

## Disclosures

* Limited data is retrieved from the in-page Gmail DOM (e.g. whether the conversation view is turned on), because ActiveInbox needs access to top.GLOBALS (only available in page), and also to listen for when Gmail refreshes data (via the XMLHttpRequest) as a trigger to tell ActiveInbox to sync with the Gmail API, which is a much more efficient approach than polling. 
** This data is extracted very securely: it's wrapped in a module/closure, it uses window.postMessage, all data is validated upon arrival in the content script, and the naming conventions of functions remind our devs to be very distrustful of the data. (See #REAL_DOM_COMM in the code). 
* ActiveInbox uses Gmail's internal API to request the modern Gmail ID for a message with a particular RFC822 ID. This is necessary because the legacy IDs in the Gmail API fail to navigate the user to the correct Gmail conversation in 20% of cases, causing a Gmail warning of 'Conversation not found', which puts Gmail is a crash-state that requires a browser refresh to start working again. (See glGmailPrivateAPI.js). 
** To secure this, it's explicitly declared in a function named getDecimalGmailIdsFromRfc822IdsForGmailNavigation, in a module. 


## Upcoming Plans

* Working on our own "PortsGuard" structure, so only a single content script module, created at run time, can communicate with a given background page module. I.e. so no malicious code can send requests to a background page, only trusted ActiveInbox code. 
* Enable Cross-Account Protection (https://developers.google.com/identity/risc), so we alert customers, or shut down OAuth approvals, if anything suspicious occurs. 
* Gradually replacing all ActiveInbox code with ES6, that's easier to grasp and audit. We've fully completed all critical aspects to TypeScript ES6, wrapped into secure modules with minimal export interfaces, with Yup schema validation (Authentication, Gmail/GCal/ActiveInbox API). (ActiveInbox is a 13 year old codebae with plenty of legacy, so this is a time-consuming ongoing refactor).
