
window.serverCommunicationBackground.googleAuthPorts();
window.serverCommunicationBackground.gmailPorts();
window.serverCommunicationBackground.gCalPorts();
window.serverCommunicationBackground.activeInboxAPIPorts();
window.serverCommunicationBackground.activeInboxStatsAPIPorts();



function debounce(func, wait, immediate) {
    var timeout;
    return function() {
        var context = this, args = arguments;
        var later = function() {
            timeout = null;
            if (!immediate) func.apply(context, args);
        };
        var callNow = immediate && !timeout;
        clearTimeout(timeout);
        timeout = setTimeout(later, wait);
        if (callNow) func.apply(context, args);
    };
};


var resetPrefs = function() { // Helper function for chrome debugging
    for( var pref in localStorage ) {
        console.log("delete "+pref);
        localStorage[pref] = null;
        delete localStorage[pref];
    }
    console.log("completed");
}

// Set up object definitions:


var PrefData = function() {

    this._prefDataPorts = [];

    this.prefs = null;
    this.pluginVersions = null;
    this.defaults = null;

};
PrefData.prototype = {

    /**
     * Connect a new port (client)
     * @function {} ?
     * @param {Port} ?
     */
    connectPort: function(port) {
        var pdp = new PrefDataPort(this, port);
        this._prefDataPorts.push(pdp);
    },

    /**
     * Disconnect port (and containing object)
     * @function {} ?
     * @param {Port} port
     */
    disconnectPort: function(port) {
        for( var i = 0; i < this._prefDataPorts.length; i++ ) {
            if( this._prefDataPorts.port===port ) {
                this._prefDataPorts.splice(i, 1);
                break;
            }
        }
    },


    /**
     * Load prefs (if necessary) then broadcast
     * @function {}
     * @param {Port} [port] If provided, just broadcast to one
     */
    load: function(port) {
        if( this.prefs===null ) {
            // Read from store
            try {
                this.prefs = giJson.decode(localStorage['extensions.gtdinbox.account_prefs'] || "{}");
            } catch(e) {
                giLogger.warn(e, 'background.html', 'PrefData.load this.prefs');
                this.prefs = {};
            }
            try {
                this.defaults = giJson.decode(localStorage['extensions.gtdinbox.account_defaults'] || "{}");
            } catch(e) {
                giLogger.warn(e, 'background.html', 'PrefData.load this.defaults');
                this.defaults = {};
            }
            try {
                this.pluginVersions = giJson.decode(localStorage['extensions.gtdinbox.plugin_versions'] || "{}");
            } catch(e) {
                giLogger.warn(e, 'background.html', 'PrefData.load this.pluginVersions');
                this.pluginVersions = {};
            }
            this._broadcast(port);
        } else {
            this._broadcast(port);
        }
    },

    /**
     * Notify all ports of latest settings
     * @function {}
     * @param {Port} [port] If provided, just broadcast to one
     * @param {Port} [exclPort] If provided, do not broadcast to just this one
     */
    _broadcast: function(port, exclPort) {
        // Broadcast:
        var msg = {'clientAction':'load', 'prefs':this.prefs, 'defaults':this.defaults, 'pluginVersions':this.pluginVersions};
        if( port ) {
            port.postMessage(msg);
        } else {
            for( var i = 0; i < this._prefDataPorts.length; i++ ) {
                if( this._prefDataPorts[i].port!=exclPort ) {
                    this._prefDataPorts[i].port.postMessage(msg);
                }
            }
        }
    },

    /**
     * Save a data object and broadcast to all
     * @function {} ?
     * @param {Object} msg Pref data
     * @param {Port} port The port that fired it
     */
    save: function(msg, port) {
        this.prefs = msg.prefs;
        this.defaults = msg.defaults;
        this.pluginVersions = msg.pluginVersions;


        // Write to store
        localStorage['extensions.gtdinbox.account_prefs'] = giJson.encode(this.prefs);
        localStorage['extensions.gtdinbox.account_defaults'] = giJson.encode(this.defaults);
        localStorage['extensions.gtdinbox.plugin_versions'] = giJson.encode(this.pluginVersions);

        // Broadcast to all except this port
        this._broadcast(null, port);
    }
};

var PrefDataPort = function(prefData, port) {
    this.port = port;
    this._prefData = prefData;

    var me = this;
    port.onMessage.addListener(function(msg){me._onMessage(msg)});

};

PrefDataPort.prototype = {
    /**
     * Notification from client
     * @function {} ?
     */
    _onMessage: function(msg) {
        if( msg.bgAction ) {
            switch( msg.bgAction ) {
            case 'load':
                // Request to load, ensure data is up to date and broadcast
                this._prefData.load(this.port);
                break;
            case 'save':
                this._prefData.save(msg, this.port);
                break;
            }
        }
    },

    /**
     * Port disconnected; remove from parent
     * @function{} ?
     */
    _onDisconnect: function() {
        this._prefData.disconnectPort(this.port);
        this._prefData = null;
        this.port = null;
    }
};
var prefData = new PrefData();














var Database = function() {
    this._connections = {};
}
Database.prototype._getDB = function(email) {
    var me = this;
    if( this._connections[email] ) {
        return Promise.resolve(this._connections[email]);
    }
    return new Promise(function(resolve, reject) {
        var request = window.indexedDB.open(email, 1);
        request.onerror = function() {
            console.error(arguments);
            reject()
        }
        request.onupgradeneeded = me._onupgradeneeded;
        request.onsuccess = function(event) {
            me._connections[email] = this.result;
            resolve(this.result);
        }
    });
}
Database.prototype._onupgradeneeded = function(event) {
    var db = event.target.result;
    db.onerror = console.error;
    var upgradeFunctions = {
        1: function() {
            db.createObjectStore('tabPacStreams', {keyPath: 'tabId'});
            db.createObjectStore('pendingActionChangesToUpload', {keyPath: 'id'});
            db.createObjectStore('settings');
        }
    };
    for( var versionNumber = event.oldVersion + 1; versionNumber <= event.newVersion; versionNumber++ ) {
        if( versionNumber in upgradeFunctions ) {
            upgradeFunctions[versionNumber]();
        }
    }
}
Database.prototype.getDataForEmail = function(email) {
    var me = this;
    const result = {
        tabPacStreams: {},
        pendingActionChangesToUpload: [],
        remoteHistoryId: null
    }
    return this.getTabPacStreams(email)
        .then(function(tabPacStreams){
            result.tabPacStreams = tabPacStreams.reduce(function(memo,tabPacStream){
                memo[tabPacStream.tabId] = tabPacStream;
                return memo;
            },{});
        })
        .then(function(){
            return me.getPendingActionChangesToUpload(email);
        })
        .then(function(pendingActionChangesToUpload){
            result.pendingActionChangesToUpload = pendingActionChangesToUpload;
        })
        .then(function(){
            return me.getSetting(email, 'remoteHistoryId');
        })
        .then(function(remoteHistoryId){
            result.remoteHistoryId = remoteHistoryId;
            return result;
        })
}
Database.prototype.setDataForEmail = function(email, tabPacStreams, pendingActionChangesToUpload, remoteHistoryId) {
    var me = this;
    return this.updateTabPacStreams(email, tabPacStreams)
        .then(function(){
            return me.updatePendingActionChangesToUpload(email, pendingActionChangesToUpload);
        })
        .then(function(){
            return me.setSetting(email, 'remoteHistoryId', remoteHistoryId);
        })
}
Database.prototype.updateTabPacStreams = function(email, tabPacStreams) {
    var updatedIds = []; // tab ids synced

    var store;
    return this._getDB(email)
        .then(function(db){
            // Update existing rows
            return new Promise(function(resolve, reject){
                store = db.transaction('tabPacStreams', 'readwrite').objectStore('tabPacStreams');
                var request = store.openCursor();
                request.onerror = reject;
                request.onsuccess = function(event){
                    var cursor = event.target.result;
                    if( cursor ) {
                        var tabId = cursor.key;
                        updatedIds.push(tabId);
                        var tabPacStream = tabPacStreams[tabId];
                        if( tabPacStream ) {
                            var a = JSON.stringify(cursor.value);
                            var b = JSON.stringify(Object.assign({tabId: tabId}, tabPacStream))
                            if( a != b ) {
                                // changed. Update
                                var req = cursor.update(Object.assign({tabId: tabId}, tabPacStream));
                                req.onsuccess = function() { cursor.continue() }
                                req.onerror = function(error) {
                                    console.warn(error);
                                    cursor.continue();
                                }
                            } else {
                                // no change. Continue
                                cursor.continue();
                            }
                        } else {
                            // tabId in database but not tabPacStreams. Delete from db
                            var req = cursor.delete();
                            req.onsuccess = function() { cursor.continue() }
                            req.onerror = function(error) {
                                console.warn(error);
                                cursor.continue();
                            }
                        }
                    } else {
                        resolve(db);
                    }
                };
            })
        })
        .then(function(db){
            return new Promise(function(resolve, reject){
                // insert new rows
                var idsToInsert = Object.keys(tabPacStreams).filter(function(tabId){ 
                    return updatedIds.includes(tabId) === false;
                })
    
                var i = 0;
                function putNext() {
                    if( i < idsToInsert.length ) {
                        var tabId = idsToInsert[i];
                        i++;
                        var tabPacStream = tabPacStreams[tabId];
                        console.log('putting', tabId);
                        store
                            .put(Object.assign({tabId: tabId}, tabPacStream))
                            .onsuccess = putNext;
                    } else {
                        resolve(db)
                    }
                }
                putNext();
            });
        });
}
Database.prototype.getTabPacStreams = function(email) {
    return this._getDB(email).then(function(db){
        return new Promise(function(resolve, reject){
            var tabPackStreams = [];
            var transaction = db.transaction('tabPacStreams');
            transaction.onerror = function(error) {
                console.warn(error);
                resolve(tabPackStreams);
            };
            transaction 
                .objectStore('tabPacStreams')
                .openCursor() 
                .onsuccess = function(event) {
                    var cursor = event.target.result;
                    if( cursor ) {
                        tabPackStreams.push(cursor.value);
                        cursor.continue();
                    } else {
                        resolve(tabPackStreams)
                    }
                }
        })
    });
}
Database.prototype.updatePendingActionChangesToUpload = function(email, pendingActionChangesToUpload) {
    var indexedRows = pendingActionChangesToUpload.reduce(function(memo, row){
        if( row.id ) {
            memo[row.id] = row;
        }
        return memo;
    },{})

    var updatedIds = [];
    var store;
    return this._getDB(email)
        .then(function(db){
            // Update existing rows
            return new Promise(function(resolve, reject){
                store = db.transaction('pendingActionChangesToUpload').objectStore('pendingActionChangesToUpload');
                var request = store.openCursor();
                request.onerror = reject;
                request.onsuccess = function(event) {
                    var cursor = event.target.result;
                    if( cursor ) {
                        var id = cursor.key;
                        updatedIds.push(id);
                        var newValue = indexedRows[id];
                        if( newValue ) {
                            var a = JSON.stringify(cursor.value);
                            var b = JSON.stringify(newValue);
                            if( a != b ) {
                                // Changed. Update
                                var req = cursor.update(newValue);
                                req.onsuccess = function() { cursor.continue() }
                                req.onerror = function(error) {
                                    console.warn(error);
                                    cursor.continue();
                                }
                            } else {
                                // No change. Continue
                                cursor.continue();
                            }
                        } else {
                            // id in database but not in pendingActionChangesToUpload
                            // Delete from db
                            var req = cursor.delete();
                            req.onsuccess = function() { cursor.continue() }
                            req.onerror = function(error) {
                                console.warn(error);
                                cursor.continue();
                            }
                        }
                    } else {
                        resolve(db)
                    }
                }
            });
        })
        .then(function(db){
            // Insert new rows
            return new Promise(function(resolve, reject){
                var idsToInsert = Object.keys(indexedRows).filter(function(id){
                    return updatedIds.includes(id) === false;
                });
                var i = 0;
                function putNext() {
                    if( i < idsToInsert.length ) {
                        var id = idsToInsert[i];
                        i++;
                        var newValue = indexedRows[id];
                        store.put(newValue).onsuccess = putNext
                    } else {
                        resolve()
                    }
                }
                putNext();
            })
        })
}
Database.prototype.getPendingActionChangesToUpload = function(email) {
    return this._getDB(email).then(function(db){
        return new Promise(function(resolve, reject){
            var transaction = db.transaction('pendingActionChangesToUpload');
            transaction.onerror = reject;
            transaction
                .objectStore('pendingActionChangesToUpload')
                .getAll()
                .onsuccess = function(event) {
                    resolve(event.target.result);
                }
        })
    });
}
Database.prototype.getSetting = function(email, key) {
    return this._getDB(email).then(function(db){
        return new Promise(function(resolve, reject){
            var transaction = db.transaction('settings');
            transaction.onerror = reject;
            transaction
                .objectStore('settings')
                .get(key)
                .onsuccess = function(event) {
                    resolve(event.target.result)
                }
        });
    });
}
Database.prototype.setSetting = function(email, key, value) {
    return this._getDB(email).then(function(db){
        return new Promise(function(resolve, reject){
            var transaction = db.transaction('settings', 'readwrite');
            transaction
                .objectStore('settings')
                .put(value, key)
                .onsuccess = resolve
        });
    });
}







// See glRsgActionsCreator for the design intent behind PendingActionChangesLocalServer (it's basically to abstract complexity of multiple tabs)
var PendingActionChangesLocalServer = function(database) {
    this._data = {};
    this._database = database;
    this._loadingDataPromise;

    // It's very costly to save (about 2s), but because we keep each email in memory, we don't need it to be up to date on the hard drive, so can delay it. 
    this._saveDataDebounced = debounce(this._saveData, 500);
}
PendingActionChangesLocalServer.prototype._loadData = function(email) {
    var me = this;
    if( this._loadedData ){ 
        // We want this._data to be set once, then kept as central state. The other functions expect it. 
        return Promise.resolve();
    }
    if( this._loadingDataPromise ) {
        return this._loadingDataPromise;
    }
    console.log("Reading in data");

    this._loadingDataPromise = me._database.getDataForEmail(email).then(function(data){
        me._data[email] = data;
        me._loadedData = true;
        me._loadingDataPromise = null;
        console.log('getPendingActionsForEmail',data);
    });
    return this._loadingDataPromise;
}
PendingActionChangesLocalServer.prototype._getDataForEmail = function(email) {
    var me = this;
    return this._loadData(email)
    .then(function(){
        console.log("Loaded data", me._data[email]);
        if( !me._data[email] ) {
            console.log("Initialised data for "+email);
            me._data[email] = {
                pendingActionChangesToUpload: [],
                remoteHistoryId: null
            }
        }
        if( !me._data[email].tabPacStreams ) me._data[email].tabPacStreams = {};
        if( !me._data[email].tabPacStreams.allSinceLastRemoteReset ) me._data[email].tabPacStreams.allSinceLastRemoteReset = {pendingActionChanges: [], lastSync: 0};
        return me._data[email];
    })
}
PendingActionChangesLocalServer.prototype._saveData = function(email) {
    var me = this;
    var data = this._data[email];
    if( data ) {
        if( data.tabPacStreams && data.pendingActionChangesToUpload ) {
            return this._database.setDataForEmail(
                email, 
                data.tabPacStreams, 
                data.pendingActionChangesToUpload, 
                data.remoteHistoryId 
            )
        }
    }
    return Promise.resolve();
}

PendingActionChangesLocalServer.prototype.registerTab = function(email, tabId, callback) {
    // TODO need a way to close down streams, or they'll just grow and grow
    var me = this;
    this._getDataForEmail(email)
    .then(function(data){

        // Failsafe for unregisterTab:
        var clearBefore = (new Date()).valueOf()-(1000*60*60*9);
        for( var oldTabId in data.tabPacStreams ) {
            if( oldTabId!='allSinceLastRemoteReset' && data.tabPacStreams[oldTabId].lastSync<clearBefore  ) {
                me._deleteTab(data, oldTabId);
            }
        }

        if( data.tabPacStreams.allSinceLastRemoteReset ) {
            var allPendingActionChanges = [];
            for( var i = 0; i < data.tabPacStreams.allSinceLastRemoteReset.pendingActionChanges.length; i++ ) {
                var pac = data.tabPacStreams.allSinceLastRemoteReset.pendingActionChanges[i];
                if( pac.timestamp>clearBefore ) {
                    allPendingActionChanges.push(pac);
                }
            }
            data.tabPacStreams.allSinceLastRemoteReset.pendingActionChanges = allPendingActionChanges;
        }
        
        // Always wipe the data for the tabPacStream at the start of a session.
        // TODO Store this in IndexedDb. It can get huge. 
        data.tabPacStreams[tabId] = {pendingActionChanges: [], lastSync: 0};

        me._saveDataDebounced(email);
    })
}
PendingActionChangesLocalServer.prototype._deleteTab = function(data, tabId) {
    if( tabId && data.tabPacStreams[tabId] ) {
        data.tabPacStreams[tabId] = null;
        delete data.tabPacStreams[tabId];
    }
    if( data.tabPacStreams.allSinceLastRemoteReset ) {
        var allPendingActionChanges = [];
        for( var i = 0; i < data.tabPacStreams.allSinceLastRemoteReset.pendingActionChanges.length; i++ ) {
            var pac = data.tabPacStreams.allSinceLastRemoteReset.pendingActionChanges[i];
            if( pac.browserTabId!==tabId ) {
                allPendingActionChanges.push(pac);
            }
        }
        data.tabPacStreams.allSinceLastRemoteReset.pendingActionChanges = allPendingActionChanges;
    }
}
PendingActionChangesLocalServer.prototype.unregisterTab = function(email, tabId, callback) {
    // TODO need a way to close down streams, or they'll just grow and grow
    console.log("unregisterTab 1 ", email, tabId);
    var me = this;
    this._getDataForEmail(email)
    .then(function(data){
        console.log("unregisterTab 2 ", email, tabId);
        me._deleteTab(data, tabId);
        console.log("unregisterTab 3 ", email, tabId);
        me._saveData(email);
        console.log("unregisterTab 4 ", email, tabId);
    })
}
PendingActionChangesLocalServer.prototype._addPacToLoadedData = function(data, pendingActionChange) {
    // Add to each tab stream:
    for( var tabId in data.tabPacStreams ) {
        var tps = data.tabPacStreams[tabId];
        tps.pendingActionChanges.push(pendingActionChange);
    }
}
PendingActionChangesLocalServer.prototype.uploadPac = function(email, pendingActionChange, callback) {
    var me = this;
    console.log("Adding data via uploadPac", pendingActionChange);
    return this._getDataForEmail(email)
    .then(function(data){
        me._addPacToLoadedData(data, pendingActionChange);
        if( pendingActionChange.pendingAction.requiresSyncBetweenDevices ) {
            // TODO Use IndexedDb to store data.pendingActionChangesToUpload, not chrome storage. 
            //      We very quickly hit a QUOTA_BYTES limit, with just 31 pending actions to upload, that stopped us loading ActiveInbox. 
            //data.pendingActionChangesToUpload.push(pendingActionChange);
        }
        me._saveDataDebounced(email);

        return data.pendingActionChangesToUpload.length? me._uploadPendingActionChangesToRemote() : Promise.resolve();
    })
    .then(function(){
        callback({success: true});
    })
    .catch(function(error){
        callback({error: error});
    })
}
PendingActionChangesLocalServer.prototype.listPacsHistory = function(email, tabId, callback) {
    var me = this;
    var promiseState = {};

    var debugId = (new Date()).valueOf()+parseInt(Math.random()*10000);
    console.log("#TIMING listPacsHistory 1", debugId, (new Date()).valueOf());

    this._getDataForEmail(email)
    .then(function(data){
        console.log("#TIMING listPacsHistory 2", debugId, (new Date()).valueOf());
        promiseState.data = data;

        return promiseState.data.tabPacStreams[tabId]? Promise.resolve() : me.registerTab(email, tabId); // For some reason it can disappear, which is sad sad sad 
    })
    .then(function(){
        // Try to upload first - otherwise risk that we'll trigger a reset of pending actions without having the full replacement ready
        return promiseState.data.pendingActionChangesToUpload.length? me._uploadPendingActionChangesToRemote() : Promise.resolve();
    })
    .then(function(data){
        console.log("#TIMING listPacsHistory 3", debugId, (new Date()).valueOf());
        return me._downloadRemotePacs(email);
    })
    .then(function(xhrResponse){
        console.log("#TIMING listPacsHistory 4", debugId, (new Date()).valueOf());
        for( var i = 0; i < xhrResponse.pendingActionChanges.length; i++ ) {
            me._addPacToLoadedData(promiseState.data, xhrResponse.pendingActionChanges[i]);
        }
        return me._isIntegrityOk(promiseState.data)? Promise.resolve() : me._resetRemotePacsIntoLocal(email);
    })
    .catch(function(error){
        switch(error.type) {
            case 'pacServerUnreachable':
                // We actually want to continue - always want to return data passing through tabs. But do let the client know. 
                promiseState.pacServerUnreachable = true;
                return Promise.resolve();
            case 'historyIdInvalid':
                return me._resetRemotePacsIntoLocal(email);
            default: 
                return Promise.reject(error);
        }
    })
    .then(function(){
        console.log("#TIMING listPacsHistory 5", debugId, (new Date()).valueOf());
        if( !tabId || !promiseState.data.tabPacStreams[tabId] ) {
            debugger;
        }
        console.log("listPacsHistory 2", debugId, promiseState.data.tabPacStreams[tabId].pendingActionChanges.slice());
        var pacs = promiseState.data.tabPacStreams[tabId].pendingActionChanges.slice();
        promiseState.data.tabPacStreams[tabId].pendingActionChanges = [];
        promiseState.data.tabPacStreams[tabId].lastSync = (new Date()).valueOf();

        console.log("listPacsHistory 3", debugId, pacs, promiseState.data.tabPacStreams[tabId].pendingActionChanges.slice(), promiseState.data);

        me._saveDataDebounced(email);
        
        console.log("#TIMING listPacsHistory 6", debugId, (new Date()).valueOf());
        console.log("listPacsHistory 4", debugId, pacs, promiseState.data.tabPacStreams[tabId].pendingActionChanges.slice());
        var outcome = {pendingActionChanges: pacs};
        if( promiseState.pacServerUnreachable ) outcome.pacServerUnreachable = true;
        callback(outcome);
    
    })
    .catch(function(error){
        me._saveDataDebounced(email);
        if( error.stack ) {
            callback({error: {message: error.stack}});
        } else if( error.message ) {
            callback({error: {message: error.message}});
        } else {
            callback({error: error});
        }
    })
}


PendingActionChangesLocalServer.prototype._isIntegrityOk = function(data) {
    // If the app has been offline for a while, or perhaps another device has, actions for a thread/draft can be in the wrong order.
    //  If that happens, throw back an invalid history id. 

    var maxTimestampByThreadId = {};
    var outOfOrder = false;
    for( var i = 0; i < data.tabPacStreams.allSinceLastRemoteReset.pendingActionChanges.length; i++ ) {
        var pac = data.tabPacStreams.allSinceLastRemoteReset.pendingActionChanges[i];
        if( maxTimestampByThreadId[pac.pendingAction.aibThreadId] ) maxTimestampByThreadId[pac.pendingAction.aibThreadId] = 0;
        if( maxTimestampByThreadId[pac.pendingAction.aibThreadId]>pac.timestamp ) {
            outOfOrder = true;
            break;
        }
        maxTimestampByThreadId[pac.pendingAction.aibThreadId] = pac.timestamp;
    }
    if( !outOfOrder ) {
        // All passed ok, so clean up (Andy: I'm assuming this is ok. This code isnt' really used at present, as we don't have a server. Might need to rethink. )
        data.tabPacStreams['allSinceLastRemoteReset'] = {pendingActionChanges: [], lastSync: 0, resetAt: (new Date()).valueOf()};
    }
    return !outOfOrder;
}
PendingActionChangesLocalServer.prototype._resetRemotePacsIntoLocal = function(email) {
    // General plan: find all affected threads, and issue deletes for them all. Then add the new ones in. 

    var me = this;
    var promiseState = {};


    return this._getDataForEmail(email)
    .then(function(data){
        promiseState.data = data;
        return me._downloadRemotePacs(email);
    })
    .then(function(xhrResponse) {

        // Reset the main session item:
        promiseState.data.remoteHistoryId = xhrResponse.historyId;
        promiseState.data.tabPacStreams['allSinceLastRemoteReset'] = {pendingActionChanges: [], lastSync: 0, resetAt: (new Date()).valueOf()};

        // This will cause all previous pending actions on the server to be deleted
        me._addPacToLoadedData(promiseState.data, {
            type: 'DELETE_PENDING_ACTIONS_REQUIRING_SYNC_BETWEEN_DEVICES',
            timestamp: (new Date()).valueOf(),
            sessionId: 'PendingActionChangesLocalServer'
        });

        // Now recreate all the pending actions... we assume the server returns the entire pac history for each pending action that's still valid
        for( var i = 0; i < xhrResponse.pendingActionChanges.length; i++ ) {
            me._addPacToLoadedData(promiseState.data, xhrResponse.pendingActionChanges[i]);
        };

        me._saveDataDebounced(email);
        return Promise.resolve();
    })
}
PendingActionChangesLocalServer.prototype._downloadRemotePacs = function(email) {
    var me = this;

    return this._getDataForEmail(email)
    .then(function(data){
        
        var xhr = null;
        // TODO download from the server, or be rejected for an old history id. 
        //      Use data.remoteHistoryId & email 
        var xhrResponse = {pendingActionChanges: [], historyId: null}; // TODO replace me
        if( "todo xhrok" ) {
            data.remoteHistoryId = xhrResponse.historyId;
            me._saveDataDebounced(email)
            return Promise.resolve(xhrResponse);
        } else if( "todo server offline") {
            return Promise.reject({type: 'pacServerUnreachable'});
        } else {
            // Nb Do this after pacServerUnreachable
            return Promise.reject({type: 'historyIdInvalid'});
        }
    })
    
}
PendingActionChangesLocalServer.prototype._uploadPendingActionChangesToRemote = function(email) {
    var me = this;
    return this._getDataForEmail(email)
    .then(function(data){
        var xhr = null;
        var uploadPacs = data.pendingActionChangesToUpload.slice();
        data.pendingActionChangesToUpload = [];
        me._saveDataDebounced(email);
        return Promise.resolve({success: true}); // TODO Remove this line, replace with server
        // TODO upload to server. Use email. BEWARE: Can the data be too large? Maybe only send some of it. 
        if( "todo xhrok" ) {
            for( var i = 0; i < uploadPacs.length; i++ ) {
                uploadPacs[i].uploaded = true;
            }
            me._saveDataDebounced(email);
            return Promise.resolve({success: true});
        } else if( "todo server offline") {
            // Undo the earlier slice:
            data.pendingActionChangesToUpload = data.pendingActionChangesToUpload.concat(uploadPacs);
            me._saveData();
            return Promise.reject({success: false, error: 'pacServerUnreachable'});
        }
    })
}

/**
    TODO
    THE PACS REMOTE SERVER
    GOALS
        For things like drafts, or if we extend the system to store notes, we want to share pending actions long term across devices. 
    IMPLEMENTING
        Stores pending action changes in table. 
            It can only delete them a month after the related pending action was marked complete/cancelled. 
            We download from the server as pending action changes, in order. 

        Because we regularly sync the server while using the app, we don't expect it to fall out of sync with other devices. 
            So we implement a history model like the Gmail API. 
            But if at any point our historyId appears out of date, or we end up with pacs in the wrong time order for a given aibThread, 
                we redownload the server pacs afresh. 

        The biggest complexity, differentiating us from Gmail History API, is:
            1. We call uploadPac and then expect it to still be there when we call listPacsHistory. 
                (This is a necessary convenience for syncing in realtime between different tabs)
                This is problematic if the history id is invalid, and forces us to do a reset. (This is why we retain session state in a reset).
            2. We essentially have two servers to keep in sync: the remote one (which only a few pacs are sent to), and this one. 

        At the moment, we don't store pacs in storage, only in session. And we download at the start of the session. 
            Therefore, we can happily blend our local pacs (never intended to hit the server), and the server pacs into one stream just fine. 
            When we reset them, we discard anything in the pendingActionStream that was previously uploaded to the server (leaving just in session), 
                and just redownload from the server. 

        How to handle syncing pending action changes in the wrong order for the same thread, due to devices being offline:
            The scenario: 
                Device 1 (offline): Create Action1 for Thread1 (e.g. add label x)
                Device 2 (offline): Create Action2 for Thread1 (e.g. remove label x)
                Device 2 goes online
                Device 2 uploads 'Create Action2 for Thread1'
                Device 2 runs 'Action2 for Thread1'
                Device 2 creates 'Completion of Action2 for Thread1'
                Device 2 uploads 'Completion of Action2 for Thread1'
                Device 1 goes online
                Device 1 uploads 'Create Action1 for Thread1'
                Device 1 syncs 'Create Action2 for Thread1' and 'Completion of Action2 for Thread1'
                Device 1 runs 'Action1 for Thread1'
                Device 1 creates 'Completion of Action1 for Thread1'
                Device 1 uploads 'Completion of Action1 for Thread1'
                = Ultimately, it results in running 'remove label x' first, then 'add label x'; the total opposite of what we wanted. 
            The solution:
                The remote server, upon upload, can detect that the .timestamp of an incoming pac is out of order with everything it knows.
                Therefore it can run over them in the correct order (without changing the order given to the client), 
                    to determine what the final outcome should be,
                    and then create additional pending action changes to make sure it happens. 
                    In the example above, that would be an additional 'remove label x'. 
                But we must handle another problem: if the user applied labels via Gmail while offline, _and_ by ActiveInbox's pending actions. 
                    Because Gmail's History API doesn't include timestamps, we can't repair the correct order. 
                    In that case, when the user reconnects, or alternatively when they enter the conversation view for an affected thread, 
                        ask them to check the state they desire for the message. 

        Reducing load on the server.
            We have a structural issue, in that the need to sync pending actions between devices to a remote server is minimal (currently only for drafts). 
            But we're going to hit our server to attempt a sync on every, single, pending action (i.e. every action a user takes). 
            It's disproportionate. 
            The solution is to only sync some of the time: 
                - If we have a draft in view in the UI, we'll require to ask the server if any other device has created pending actions for it. 
                - If we send an email, we'll have to sync the server first, 
                    as we're the only device that'll have a chance to get draft id & final id, to apply any pending actions to the final id. 
                - If we're enacting pending actions in the client, and any of my pending actions have stated they requiresSyncBetweenDevices, 
                    we'll need to check the server in case another device tried enacting them first. 

        Note that the history id used here is different to the remoteHistoryId. 
            The client of this just cares about getting all changes. 
            But if we've been offline for a while, our remoteHistoryId will be behind this historyid. 
                Hence they must stay separate, so the remote pending action changes from the past can be added here, 
                and delivered to the client, as if they were new.

    MISC DESIGN CONSIDERATIONS
        What happens if we can interoperate with Gmail's offline mode? 
            In that case, we simply don't need to add those to the remote server (no .requiresSyncBetweenDevices). 
            We can let Gmail take the burden instead. 

        What happens if we can add labels to drafts?
            We don't need to add those to the remote server (no .requiresSyncBetweenDevices). 

        If Gmail makes it possible to handle both offline and drafts, we can probably drop pendingActions for add/remove labels. 
            (We wouldn't even need to sync multi-tab, as we simply wouldn't be creating the pending action data). 
            We might still want to use the mechanism for Sub Tasks & Notes though. 
*/

var database = new Database();
var pendingActionChangesLocalServer = new PendingActionChangesLocalServer(database);



























// Set up listeners:
var fBcOnConnect = (chrome.runtime && chrome.runtime.onConnect)? chrome.runtime.onConnect : chrome.extension.onConnect;
fBcOnConnect.addListener(function(port, id){
    if( port.name=='prefData' ) {
        prefData.connectPort(port);
    }
    if( port.name=='gmailApiSafeFetch' ) {
        new GmailApiSafeFetch(port);
    }
});

var lastExtensionUpdate = {};

var fBcOnMessage = chrome.runtime? chrome.runtime.onMessage : chrome.extension.onMessage;
fBcOnMessage.addListener(function(request, sender, sendResponse){
    switch( request.name ) {
        case 'getLanguage':
            //sendResponse(window.navigator.language);
            chrome.i18n.getAcceptLanguages(function(langList) {
                var lang = langList[0] || window.navigator.language;
                sendResponse(lang);
            })
            return true; // let sendResponse be async
            break;
        case 'tab.create':
            if( !window.securityVerified.background ) throw Error("Cannot tab create until security code in place");
            chrome.tabs.create({"url":request.url, "selected":request.selectTheTab}, function(tab){
                if(request.closeAfter) {
                    setTimeout(function(){
                        chrome.tabs.remove(tab.id);
                    }, request.closeAfter);
                }
            });
            sendResponse();
            break;
        case 'lastExtensionUpdate':
            sendResponse(lastExtensionUpdate);
            break;
        case 'registerTab':
            pendingActionChangesLocalServer.registerTab(request.email, request.tabId, sendResponse);
            break;
        case 'unregisterTab':
            pendingActionChangesLocalServer.unregisterTab(request.email, request.tabId, sendResponse);
            break;
        case 'uploadPac':
            pendingActionChangesLocalServer.uploadPac(request.email, request.pendingActionChange, sendResponse);
            return true; // let uploadPac be async
            break;
        case 'listPacsHistory':
            pendingActionChangesLocalServer.listPacsHistory(request.email, request.tabId, sendResponse);
            return true; // let listPacsHistory be async
            break;
        case 'aibApiCorbSafe':
            var options = request.options;
            if( /https:\/\/api([^.]+)?.activeinboxhq.com/.test(options.url) ) {
                var xhr = new XMLHttpRequest();
                xhr.open(options.method, options.url, true);
                xhr.setRequestHeader('Accept', 'application/json');
                if (options.method !== 'GET') {
                    xhr.setRequestHeader('Content-Type', 'application/x-www-form-urlencoded');
                }
                // generate the appropriate Authorization header
                xhr.setRequestHeader('Authorization', options.authHeader);
                xhr.onreadystatechange = function (evt) {
                    if( xhr.readyState===4 ) {
                        var json = null;
                        var success = false;
                        var errorType = null;
                        if ( xhr.status > 100 && xhr.status < 400 ) {
                            try { 
                                json = JSON.parse(xhr.responseText);
                            } catch(e) {}
                            // Due to sendResponse limitations, responseText might be truncated.
                            success = true;
                        } else {
                            errorType = 'http_status_error';
                        }
                        var response = {success: success, readyState: xhr.readyState, status: xhr.status, responseText: xhr.responseText, json: json, statusText: xhr.statusText};
                        if( errorType ) {
                            response.type = errorType;
                            response.request = {
                                method: options.method,
                                url: options.url
                            };
                        }
                        sendResponse(response);
                    }
                };

                xhr.send(options.body);
            } else {
                sendResponse({success: false, error: {type: 'invalid_aib_api_url'}})
            }
            return true; // Let it be async
            break;
        case 'googleCalendarCorbSafe':
            var apiRequest = {
                method: request.method,
                headers: {
                    'Authorization': 'Bearer ' + request.accessToken
                },
                body: null
            }
            if( request.method.toUpperCase()!=='GET' ) {
                apiRequest.headers['Content-Type'] = 'application/json';
                if( request.options ) apiRequest.body = JSON.stringify(request.options)
            }
            fetch('https://www.googleapis.com/calendar/v3' + request.path, apiRequest)
            .then(function(response){
                try {
                    return response.json();
                } catch(e){
                    sendResponse({success: false, error: {type: 'decode_json', message: e.message}});
                }
            })
            .then(function(json){
                sendResponse({success: true, json: json});
            })
            .catch(function(error){
                var safeErrorKeys = ['type', 'code', 'status', 'message', 'responseText'];
                var safeError = {};
                safeErrorKeys.forEach(function(key) { safeError[key] = error[key] });
                sendResponse({success: false, error: safeError});
            })
            return true; // Let it be async
            break;
        default:
            sendResponse();
            break;
    }

});

chrome.runtime.onInstalled.addListener(function(details){
    if( details.reason=='update' ) {
        lastExtensionUpdate.ts = (new Date()).valueOf();
        lastExtensionUpdate.previousVersion = details.previousVersion;
    }
});





var GmailApiSafeFetch = function(port) {
    this.port = port;
    this.lastFetchResponse = null;
    this._debugTested401 = false;

    var me = this;
    port.onMessage.addListener(function(msg){me._onMessage(msg)});

};

GmailApiSafeFetch.prototype = {
    _onMessage: function(msg) {
        var me = this;
        switch( msg.name ) {
            case 'fetch':
                if( !/https:\/\/([^.]+).googleapis.com/.test(msg.url) ) {
                    me.port.postMessage({'fetchStatus': 'error', error: {type: 'invalid_gmail_api_url'}});
                    return;
                }
                me.lastFetchResponse = null;


                if( msg.props.headers ) msg.props.headers = new Headers(msg.props.headers);
                
                me.state = {};
                me.state.url = msg.url;
                me.state.props = msg.props;

                fetch(msg.url, msg.props)
                .then(function(response){
                    /*
                    if( !me._debugTested401 ) {
                        me._debugTested401 = true;
                        return Promise.resolve()
                        .then(function(){
                            var error = {type: 'api_http_status_error', status: 401, code: 401, responseText: ""};
                            me.port.postMessage({'fetchStatus': 'error', error: error})
                        })
                    }
                    */
                    if( response.ok ) {
                        me.lastFetchResponse = response;
                        me.responseIsStreamable = !!response.body;
                        me.responseBoundary = response.headers.get('content-type').split('boundary=')[1];
                        me.responseNextMultipleResults = null;
                        me.port.postMessage({'fetchStatus': 'success'})
                    } else {
                        return (response.text? response.text() : Promise.resolve())
                        .then(function(txt){
                            if( response.status===500 ) debugger;
                            var error = {type: 'api_http_status_error', status: response.status, code: response.status, responseText: txt};
                            me.port.postMessage({'fetchStatus': 'error', error: error})
                        })
                    }
                })
                .catch(function(error){
                    var safeErrorKeys = ['type', 'code', 'status', 'message', 'responseText'];
                    var safeError = {};
                    safeErrorKeys.forEach(function(key) { safeError[key] = error[key] });
                    me.port.postMessage({'fetchStatus': 'error', error: safeError})
                })

                break;
            case 'json':
                me.lastFetchResponse.json()
                .then(function(data) {
                    try {
                        me.port.postMessage({json: data});
                    } catch(e) {
                        debugger;
                        var charLength = 0; 
                        try {
                            charLength = JSON.stringify(data).length;
                        }catch(e2){}
                        me.port.postMessage({'fetchStatus': 'error', error: {type: 'post_transport_error', charLength: charLength, message: e.message}});
                        console.log('#POST_TRANSPORT_ERROR json', 'e: ', e, 'result: ', result, 'data: ', data);
                    }
                })
                break;
            case 'nextMultipartResults':
                var multipartResultsBuffer = [];
                if( !me.postNextMultipleResults ) {
                    me.postNextMultipleResults = function(streamDone){
                        var multipartResultsToSend = [];
                        var done = streamDone && multipartResultsBuffer.length===0;
                        try {
                            if( !done ) {
                                // Only send what we can transmit safely to the background page, which is variable depending on computer power 
                                var MAX_SIZE = (1024*256);
                                var size = 0;
                                var itemSize = 0;
                                while(multipartResultsBuffer.length){
                                    itemSize = me._byteCount(multipartResultsBuffer[0]);
                                    
                                    
                                    size += itemSize;
                                    multipartResultsToSend.push(multipartResultsBuffer[0]);
                                    multipartResultsBuffer.shift();
                                    if( size>MAX_SIZE || multipartResultsToSend.length>=25 ) {
                                        // THIS IS INTENTIONALLY AFTER THE PUSH. WE WANT IT TO PUT AT LEAST ONE ON. 
                                        console.log("BREAK BUFFER", multipartResultsToSend.length, size);
                                        break;
                                    }
                                }
                            }
                            //console.log("SEND", multipartResultsToSend.length);
                            me.port.postMessage({done: done, multipartResults: multipartResultsToSend});
                        } catch(e) {
                            debugger;
                            var charLength = 0; 
                            try {
                                charLength = JSON.stringify(multipartResultsToSend).length;
                            }catch(e2){}
                            me.port.postMessage({'fetchStatus': 'error', error: {type: 'post_transport_error', charLength: charLength, message: e.message}});
                            console.log('#POST_TRANSPORT_ERROR nextMultipartResults', 'e: ', e,  'multipartResultsBuffer: ', multipartResultsBuffer, 'multipartResultsToSend: ', multipartResultsToSend);
                        }
                    }
                }
                if( me.responseIsStreamable ) {
                    if( !me.responseNextMultipleResults ) {
                        var reader = me.lastFetchResponse.body.getReader();
                        var decoder = new TextDecoder();
                        var bodyPart = '';
                        var streamDone = false;
                        var count = 0;
                        me.responseNextMultipleResults = function() {
                            (streamDone? Promise.resolve() : reader.read())
                            .then(function(result) {
                                if( result ) {
                                    bodyPart += decoder.decode(result.value || new Uint8Array, {
                                        stream: !result.done
                                    });
                                    var multipartResults = [];
                                    if( result.done ) {
                                        streamDone = true;
                                    }
                                    if( bodyPart ) {
                                        var multipartResults = bodyPart.split('--'+me.responseBoundary);

                                        multipartResults = multipartResults.filter(function(multipartResult){
                                            return !!multipartResult && multipartResult !== '--\r\n';
                                        });

                                        // the last part will probably not be complete. So stash it until the next
                                        // chunk comes in
                                        let malformedLastResult = !multipartResults[multipartResults.length - 1] || !multipartResults[multipartResults.length - 1].lastIndexOf;
                                        if( malformedLastResult ) {
                                            let mrClone = (multipartResults && multipartResults.slice)? multipartResults.slice() : multipartResults;
                                            console.warn("Major issue, where the last mutipartResults was malformed. Possibly just no results?", mrClone);
                                        }
                                        if( !malformedLastResult && multipartResults[multipartResults.length - 1].lastIndexOf('\n}') == -1 ){
                                            bodyPart = multipartResults.pop(); // This will concat it for the next run. 
                                        } else {
                                            bodyPart = '';
                                        }

                                        multipartResultsBuffer = multipartResultsBuffer.concat(multipartResults);
                                    }
                                }


                                me.postNextMultipleResults(streamDone);
                            });
                        }
                    }
                    me.responseNextMultipleResults();
                } else {
                    // WARNING this loads all multipart items into memory and processes
                    // them in one batch. We only do this because Firefox 2018 does not 
                    // currently support ReadableStream. This should be avoided if at all
                    // possible
                    if( !me.responseNextMultipleResults ) {
                        var splitBoundaries = false;
                        me.responseNextMultipleResults = function() {
                            (hasSplitBoundaries? Promise.resolve() : me.lastFetchResponse.text())
                            .then(function(responseText){
                                if( !hasSplitBoundaries ) {
                                    hasSplitBoundaries = true;
                                    multipartResultsBuffer = responseText.split('--'+boundary);
                                }
                                me.postNextMultipleResults(true);
                            })
                        }
                    }
                    me.responseNextMultipleResults();
                }
                break;
        }
    },

    _byteCount(s) {
        //return (1024*1024*12);
        return encodeURI(s).split(/%..|./).length - 1;
    },

    /**
     * Port disconnected; remove from parent
     * @function{} ?
     */
    _onDisconnect: function() {
        this.port = null;
    }
};



